@@ -546,16 +546,22 @@ fi\n",
   if (timestamp_mode)
     {
       const char *file = "$$.touch";
-      const char *stamp = "1231235999";
+      const char *stamp1 = "200112312359.59";
+      const char *stamp2 = "123123592001.59";
+      const char *stamp2tr = "123123592001.5"; /* old SysV 14-char limit */
+      const char *stamp3 = "1231235901";
 
-      /* Set the shell variable shar_touch to `touch' if the touch
-	 program is proven able to restore dates.  Otherwise, set
-	 shar_touch to `:' and issue a warning.  */
+      /* Set the shell variable shar_touch to an appropriate invocation
+	 of `touch' if the touch program is proven able to restore dates.
+	 Otherwise, set shar_touch to `:' and issue a warning.  */
 
       fprintf (output, "\
-touch -am %s %s >/dev/null 2>&1\n\
-if test ! -f %s && test -f %s; then\n\
-  shar_touch=touch\n\
+if touch -am -t %s %s >/dev/null 2>&1 && test ! -f %s -a -f %s; then\n\
+  shar_touch='touch -am -t $1$2$3$4$5$6.$7 \"$8\"'\n\
+elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s -a ! -f %s -a -f %s; then\n\
+  shar_touch='touch -am $3$4$5$6$1$2.$7 \"$8\"'\n\
+elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s -a -f %s; then\n\
+  shar_touch='touch -am $3$4$5$6$2 \"$8\"'\n\
 else\n\
   shar_touch=:\n\
   echo\n\
@@ -563,14 +569,16 @@ else\n\
   $echo \"%s\"\n\
   echo\n\
 fi\n\
-rm -f %s %s\n\
+rm -f %s %s %s %s %s\n\
 #\n",
-	       stamp, file, stamp, file,
+	       stamp1, file, stamp1, file,
+	       stamp2, file, stamp2, stamp2tr, file,
+	       stamp3, file, stamp3, file,
 	       N_("\
 WARNING: not restoring timestamps.  Consider getting and"),
 	       N_("\
 installing GNU \\`touch', distributed in GNU File Utilities..."),
-	       stamp, file);
+	       stamp1, stamp2, stamp2tr, stamp3, file);
     }
 
   if (!split_file_mode || part_number == 1)
@@ -1489,10 +1497,12 @@ else\n",
 
       restore_time = localtime (&struct_stat.st_mtime);
       fprintf (output, "\
-  $shar_touch -am %02d%02d%02d%02d%02d '%s' &&\n",
+  (set %02d %02d %02d %02d %02d %02d %02d '%s'; eval \"$shar_touch\") &&\n",
+	       (restore_time->tm_year + 1900) / 100,
+	       (restore_time->tm_year + 1900) % 100,
 	       restore_time->tm_mon + 1, restore_time->tm_mday,
 	       restore_time->tm_hour, restore_time->tm_min,
-	       restore_time->tm_year, restore_name);
+	       restore_time->tm_sec, restore_name);
     }
 
   if (vanilla_operation_mode)
